from contextvars import ContextVar

correlation_id: ContextVar = ContextVar('correlation_id', default=None)
celery_parent_id: ContextVar = ContextVar('celery_parent', default=None)
celery_current_id: ContextVar = ContextVar('celery_current', default=None)



from asgi_correlation_id.log_filters import CeleryTracingIds, CorrelationId
from asgi_correlation_id.middleware import CorrelationIdMiddleware

__all__ = ('CorrelationId', 'CeleryTracingIds', 'CorrelationIdMiddleware')



from typing import Any
from uuid import uuid4

from celery import Task
from celery.signals import before_task_publish, task_postrun, task_prerun

from asgi_correlation_id.context import celery_current_id, celery_parent_id, correlation_id
from asgi_correlation_id.sentry import set_transaction_id


def configure_celery(
    uuid_length: int = 32,
    log_parent: bool = True,
    parent_header: str = 'CELERY_PARENT_ID',
    correlation_id_header: str = 'CORRELATION_ID',
) -> None:
    """
    Configure Celery event hooks.

    Passes correlation IDs from parent processes to child processes in a Celery context.

    This means a correlation ID can be transferred from a request to a worker, or from a worker to another worker.

    For workers executing scheduled tasks, a correlation ID is generated for each new task.
    """

    @before_task_publish.connect
    def publish_task_from_worker_or_request(headers: dict, **kwargs: Any) -> None:
        """
        Transfer correlation ID from request thread to Celery worker, by adding
        it as a header.

        This way we're able to correlated work executed by Celery workers, back
        to the originating request (if there was one).
        """
        id_value = correlation_id.get()
        headers[correlation_id_header] = id_value

        if log_parent:
            current = celery_current_id.get()
            if current:
                headers[parent_header] = current

    @task_prerun.connect
    def worker_prerun(task: Task, **kwargs: Any) -> None:
        """
        Set request ID, current ID, and parent ID if state was transferred.
        """
        id_value = task.request.get(correlation_id_header)

        if id_value:
            correlation_id.set(id_value)
            set_transaction_id(id_value)
        else:
            generated_correlation_id = uuid4().hex[:uuid_length]
            correlation_id.set(generated_correlation_id)
            set_transaction_id(generated_correlation_id)

        if log_parent:
            origin = task.request.get(parent_header)

            if origin:
                celery_parent_id.set(origin)

            generated_current_id = uuid4().hex[:uuid_length]
            celery_current_id.set(generated_current_id)

    @task_postrun.connect
    def clean_up(_: Task, **kwargs: Any) -> None:
        """
        Clear the context vars, to avoid re-using values in the next task.
        """
        correlation_id.set(None)
        if log_parent:
            celery_current_id.set(None)
            celery_parent_id.set(None)



import logging
from dataclasses import dataclass
from uuid import UUID, uuid4

from starlette.datastructures import Headers
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from asgi_correlation_id.context import correlation_id

logger = logging.getLogger('asgi_correlation_id')


def is_valid_uuid(uuid_: str) -> bool:
    """
    Check whether a string is a uuid.
    """
    try:
        return bool(UUID(uuid_, version=4))
    except ValueError:
        return False


@dataclass()
class CorrelationIdMiddleware:
    app: ASGIApp
    uuid_length: int = 32
    validate_guid: bool = True
    correlation_id_header_name: str = 'Correlation-ID'

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope['type'] != 'http':
            return await self.app(scope, receive, send)

        header_value = Headers(scope=scope).get(self.correlation_id_header_name.lower())

        if not header_value or (self.validate_guid and not is_valid_uuid(header_value)):
            logger.warning('Generating new UUID after receiving invalid header value: %s', header_value)
            correlation_id.set(uuid4().hex[: self.uuid_length])
        else:
            correlation_id.set(header_value[: self.uuid_length])

        async def handle_outgoing_request(message: Message) -> None:
            if message['type'] == 'http.response.start':
                headers = {k.decode(): v.decode() for (k, v) in message['headers']}
                headers[self.correlation_id_header_name] = correlation_id.get()
                headers['Access-Control-Expose-Headers'] = self.correlation_id_header_name
                response_headers = Headers(headers=headers)
                message['headers'] = response_headers.raw

            await send(message)

        return await self.app(scope, receive, handle_outgoing_request)



"""
When the Sentry SDK is installed, we want to push a request's
correlation ID, or a Celery tasks tracing ID as a transaction_id.

We could consider guarding this behind a setting if requested.
"""
try:
    from sentry_sdk import configure_scope

    def set_transaction_id(correlation_id: str) -> None:
        """
        Push the correlation ID as Sentry's transaction ID.

        The transaction ID will appear in the Sentry UI and can be used to correlate
        logs to events.
        """
        with configure_scope() as scope:
            scope.set_tag('transaction_id', correlation_id)


except ImportError:

    def set_transaction_id(correlation_id: str) -> None:
        """
        Do nothing if Sentry isn't installed.
        """
        pass



from logging import Filter, LogRecord

from asgi_correlation_id.context import celery_current_id, celery_parent_id
from asgi_correlation_id.middleware import correlation_id


class CorrelationId(Filter):
    def filter(self, record: LogRecord) -> bool:
        """
        Append a 'correlation_id' value to the log-record.
        """
        record.correlation_id = correlation_id.get()  # type: ignore[attr-defined]
        return True


class CeleryTracingIds(Filter):
    def filter(self, record: LogRecord) -> bool:
        """
        Append 'celery_parent_id' and 'celery_current_id' to log-record.

        Celery parent ID is the tracing ID of the process that spawned the
        current process, and celery current is the current process' tracing ID.
        In other words, if a worker sent a task to be executed by the worker pool,
        that celery worker's current tracing ID would become the next task worker's
        parent tracing ID.
        """
        record.celery_parent_id = celery_parent_id.get()  # type: ignore[attr-defined]
        record.celery_current_id = celery_current_id.get()  # type: ignore[attr-defined]
        return True



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asgi_correlation_id']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'asgi-correlation-id',
    'version': '0.1.1',
    'description': 'Middleware that propagates HTTP header correlation IDs to project logs',
    'long_description': '[![test](https://github.com/snok/asgi-correlation-id/actions/workflows/test.yml/badge.svg)](https://github.com/snok/asgi-correlation-id/actions/workflows/test.yml)\n[![codecov](https://codecov.io/gh/snok/asgi-correlation-id/branch/master/graph/badge.svg?token=1aXlWPm2gb)](https://codecov.io/gh/snok/asgi-correlation-id)\n[![pypi](https://pypi.org/pypi/asgi-correlation-id)](https://github.com/snok/asgi-correlation-id)\n\n\n# ASGI Correlation ID middleware\n\nThis is a rewrite of [django-guid](django-guid) for ASGI apps.\n',
    'author': 'Jonas Krüger Svensson',
    'author_email': 'jonas-ks@hotmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/snok/asgi-correlation-id',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
